#!/usr/bin/env python3

import time
import argparse
import hjson
from typing import Any, Dict

from c3.utils.utils import num3str
from rich.console import Console
from rich.table import Table


def show_table(log: Dict[str, Any], console: Console) -> None:
    """Generate a rich table from an optimization status and display it on the console.

    Parameters
    ----------
    log : Dict
        Dictionary read from a json log file containing a c3-toolset optimization status.
    console : Console
        Rich console for output.
    """
    if log:
        opt_map = log["opt_map"]
        optim_status = log["optim_status"]
        units = log["units"]
        params = optim_status["params"]
        grads = optim_status["gradient"]

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Parameter")
        table.add_column("Value", justify="right")
        table.add_column("Gradient", justify="right")
        for ii, equiv_ids in enumerate(opt_map):
            par = params[ii]
            grad = grads[ii]
            par = num3str(par)
            grad = num3str(grad)
            par_id = equiv_ids[0]
            table.add_row(par_id, par + units[ii], grad + units[ii])
            if len(equiv_ids) > 1:
                for par_id in equiv_ids[1:]:
                    table.add_row(par_id, "''", "''")

        console.clear()
        print(
            f"Optimization reached {optim_status['goal']:0.3g} at {optim_status['time']}\n"
        )
        console.print(table)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("log_file")
    parser.add_argument(
        "-w",
        "--watch",
        type=int,
        default=0,
        help="Update the table every WATCH seconds.",
    )
    args = parser.parse_args()

    try:
        with open(args.log_file) as file:
            log = hjson.load(file)
        console = Console()
        if args.watch:
            while True:
                show_table(log, console)
                time.sleep(args.watch)
        else:
            show_table(log, console)
    except FileNotFoundError:
        print("Logfile not found. Quiting...")



import numpy as np
import pytest

from numpy.testing import assert_array_almost_equal as almost_equal
from c3.libraries.fidelities import unitary_infid, average_infid
from c3.libraries.constants import GATES


X = GATES["rxp"]
Y = GATES["ryp"]
Id = GATES["id"]


@pytest.mark.unit
def test_unitary_infid_1() -> None:
    """Testing that a matrix has no error with itself."""
    almost_equal(unitary_infid(X, X, dims=[2]), 0)


@pytest.mark.unit
def test_unitary_infid_2() -> None:
    """Testing that X and Y have maximal error."""
    assert unitary_infid(X, Y, dims=[2]) == 1


@pytest.mark.unit
def test_unitary_infid_3() -> None:
    """Testing that a matrix has no error with itself."""
    actual = np.kron(X, Id)
    almost_equal(unitary_infid(actual, actual, index=[0, 1], dims=[2, 2]), 0)


@pytest.mark.unit
def test_unitary_infid_projection() -> None:
    """Testing only one subspace."""
    actual = np.kron(X, Id)
    almost_equal(unitary_infid(X, actual, index=[0], dims=[2, 2]), 0)


@pytest.mark.unit
def test_unitary_infid_projection_2() -> None:
    """Testing another subspace."""
    actual = np.kron(Id, X)
    almost_equal(unitary_infid(X, actual, index=[1], dims=[2, 2]), 0)


@pytest.mark.unit
def test_unitary_infid_projection_3() -> None:
    """Testing higher levels."""
    actual = np.array(
        [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 0]],
    )
    almost_equal(unitary_infid(ideal=X, actual=actual, index=[0], dims=[3]), 0)


@pytest.mark.unit
def test_unitary_infid_projection_4() -> None:
    """Testing higher levels."""
    actual = np.array(
        [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 34345j]],
    )
    almost_equal(unitary_infid(ideal=X, actual=actual, index=[0], dims=[3]), 0)


@pytest.mark.unit
def test_unitary_infid_projection_5() -> None:
    """Testing higher levels and subspaces."""
    actual = np.kron(
        np.array(
            [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 34345j]],
        ),
        Id,
    )
    almost_equal(unitary_infid(ideal=X, actual=actual, index=[0], dims=[3, 2]), 0)


@pytest.mark.unit
def test_average_infid_1() -> None:
    """Testing that a matrix has no error with itself."""
    almost_equal(average_infid(X, X), 0)


@pytest.mark.unit
def test_average_infid_2() -> None:
    """Testing that X and Y have maximal error."""
    almost_equal(average_infid(X, Y), np.array(2.0 / 3))


@pytest.mark.unit
def test_average_infid_projection() -> None:
    """Testing only one subspace."""
    actual = np.kron(X, Id)
    almost_equal(average_infid(X, actual, index=[0], dims=[2, 2]), 0)


@pytest.mark.unit
def test_average_infid_projection_2() -> None:
    """Testing another subspace."""
    actual = np.kron(Id, X)
    almost_equal(average_infid(X, actual, index=[1], dims=[2, 2]), 0)


@pytest.mark.unit
def test_average_infid_projection_3() -> None:
    """Testing higher levels."""
    actual = np.array(
        [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 0]],
    )
    almost_equal(average_infid(ideal=X, actual=actual, index=[0], dims=[3]), 0)


@pytest.mark.unit
def test_average_infid_projection_4() -> None:
    """Testing higher levels."""
    actual = np.array(
        [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 34345j]],
    )
    almost_equal(average_infid(ideal=X, actual=actual, index=[0], dims=[3]), 0)


@pytest.mark.unit
def test_average_infid_projection_5() -> None:
    """Testing higher levels and subspaces."""
    actual = np.kron(
        np.array(
            [[0 + 0j, 1, 0], [1, 0, 0], [0, 0, 34345j]],
        ),
        Id,
    )
    almost_equal(average_infid(ideal=X, actual=actual, index=[0], dims=[3, 2]), 0)



from qiskit.providers import JobV1 as Job
from qiskit.providers.jobstatus import JobStatus
from qiskit.result import Result


class C3Job(Job):
    """C3Job class

    Parameters
    ----------
    Job : JobV1
        Inherits JobV1 from qiskit.providers

    """

    _async = False

    def __init__(self, backend, job_id, result):
        super().__init__(backend, job_id)
        self._result = result

    def submit(self) -> None:
        """Submit a job to the simulator"""
        return

    def result(self) -> Result:
        """Return the result of the job

        Returns
        -------
        qiskit.Result
            Result of the job simulation
        """
        return self._result

    def status(self) -> JobStatus:
        """Return job status

        Returns
        -------
        qiskit.providers.JobStatus
            Status of the job
        """
        return JobStatus.DONE



"""Module for Sensitivity Analysis. This allows the sweeping of the goal
function in a given range of parameters to ascertain whether the dataset
being used is sensitive to changes in the parameters of interest"""

from copy import deepcopy
import os
from typing import Any, Dict, List, Tuple
from c3.optimizers.modellearning import ModelLearning
from c3.parametermap import ParameterMap
from c3.utils.utils import flatten


class Sensitivity(ModelLearning):
    """Class for Sensitivity Analysis, subclassed from Model Learning

    Parameters
    ----------
    sampling : str
        Name of the sampling method from library
    batch_sizes : Dict[str, int]
        Number of points to select from the dataset
    pmap : ParameterMap
        Model parameter map
    datafiles : Dict[str, str]
        The datafiles for each of the learning datasets
    state_labels : Dict[str, List[Any]]
        The labels for the excited states of the system
    sweep_map : List[List[List[str]]]
        Map of variables to be swept in exp_opt_map format
    sweep_bounds : List[List[int]]
        List of upper and lower bounds for each sweeping variable
    algorithm : str
        Name of the sweeping algorithm from the library
    estimator : str, optional
        Name of estimator method from library, by default None
    estimator_list : List[str], optional
        List of different estimators to be used, by default None
    dir_path : str, optional
        Path to save sensitivity logs, by default None
    run_name : str, optional
        Name of the experiment run, by default None
    options : dict, optional
        Options for the sweeping algorithm, by default {}

    Raises
    ------
    NotImplementedError
        When trying to set the estimator or estimator_list
    """

    def __init__(
        self,
        sampling: str,
        batch_sizes: Dict[str, int],
        pmap: ParameterMap,
        datafiles: Dict[str, str],
        state_labels: Dict[str, List[Any]],
        sweep_map: List[List[Tuple[str]]],
        sweep_bounds: List[List[int]],
        algorithm: str,
        estimator: str = None,
        estimator_list: List[str] = None,
        dir_path: str = None,
        run_name: str = None,
        options={},
    ) -> None:

        super().__init__(
            sampling,
            batch_sizes,
            pmap,
            datafiles,
            dir_path,
            estimator,
            state_labels=state_labels,
            algorithm=algorithm,
            run_name=run_name,
            options=options,
        )
        if estimator_list:
            raise NotImplementedError(
                "C3:ERROR: Estimator Lists are currently not supported."
                "Only the standard logarithmic likelihood can be used at the moment."
                "Please remove this setting."
            )
        self.sweep_map = sweep_map  # variables to be swept
        self.pmap.opt_map = [
            sweep_map[0]
        ]  # set the opt_map to the first sweep variable
        self.sweep_bounds = sweep_bounds  # bounds for sweeping
        self.sweep_end: List[
            Dict[Any, Any]
        ] = list()  # list for storing the params and goals at the end of the sweep
        self.scaling = False  # interoperability with model learning which uses scaling
        self.logname = "sensitivity.log"  # shared log_setup requires logname
        self.logdir_list: List[
            str
        ] = list()  # list of logdirs for all the different sweeps
        self.run = self.sensitivity  # alias for legacy method

    def sensitivity(self):
        """
        Run the sensitivity analysis.
        """
        for ii in range(len(self.sweep_map)):
            self.pmap.opt_map = [self.sweep_map[ii]]
            self.options["bounds"] = [self.sweep_bounds[ii]]
            print(f"C3:STATUS:Sweeping {self.pmap.opt_map}: {self.sweep_bounds[ii]}")
            self.log_setup()
            self.start_log()
            print(f"C3:STATUS:Saving as: {os.path.abspath(self.logdir + self.logname)}")
            x_init = [self.options["init_point"][ii]]
            try:
                self.algorithm(
                    x_init,
                    fun=self.fct_to_min,
                    fun_grad=self.fct_to_min_autograd,
                    grad_lookup=self.lookup_gradient,
                    options=self.options,
                )
            except KeyboardInterrupt:
                pass
            temp_param_name = "".join(flatten(self.pmap.opt_map))
            self.sweep_end.append({temp_param_name: deepcopy(self.optim_status)})
            self.logdir_list.append(deepcopy(self.logdir))



"""
testing module for Model class
"""
import pickle

import pytest
import numpy as np
from c3.c3objs import Quantity
from c3.libraries.chip import Qubit, Coupling, Drive
from c3.libraries.tasks import InitialiseGround, ConfusionMatrix
from c3.model import Model
import c3.libraries.hamiltonians as hamiltonians
from c3.parametermap import ParameterMap

qubit_lvls = 3
freq_q1 = 5e9
anhar_q1 = -210e6
t1_q1 = 27e-6
t2star_q1 = 39e-6
qubit_temp = 50e-3

q1 = Qubit(
    name="Q1",
    desc="Qubit 1",
    freq=Quantity(value=freq_q1, min_val=4.995e9, max_val=5.005e9, unit="Hz 2pi"),
    anhar=Quantity(value=anhar_q1, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
    hilbert_dim=qubit_lvls,
    t1=Quantity(value=t1_q1, min_val=1e-6, max_val=90e-6, unit="s"),
    t2star=Quantity(value=t2star_q1, min_val=10e-6, max_val=90e-3, unit="s"),
    temp=Quantity(value=qubit_temp, min_val=0.0, max_val=0.12, unit="K"),
)

freq_q2 = 5.6e9
anhar_q2 = -240e6
t1_q2 = 23e-6
t2star_q2 = 31e-6
q2 = Qubit(
    name="Q2",
    desc="Qubit 2",
    freq=Quantity(value=freq_q2, min_val=5.595e9, max_val=5.605e9, unit="Hz 2pi"),
    anhar=Quantity(value=anhar_q2, min_val=-380e6, max_val=-120e6, unit="Hz 2pi"),
    hilbert_dim=qubit_lvls,
    t1=Quantity(value=t1_q2, min_val=1e-6, max_val=90e-6, unit="s"),
    t2star=Quantity(value=t2star_q2, min_val=10e-6, max_val=90e-6, unit="s"),
    temp=Quantity(value=qubit_temp, min_val=0.0, max_val=0.12, unit="K"),
)

coupling_strength = 20e6
q1q2 = Coupling(
    name="Q1-Q2",
    desc="coupling",
    comment="Coupling qubit 1 to qubit 2",
    connected=["Q1", "Q2"],
    strength=Quantity(
        value=coupling_strength, min_val=-1 * 1e3, max_val=200e6, unit="Hz 2pi"
    ),
    hamiltonian_func=hamiltonians.int_XX,
)

drive = Drive(
    name="d1",
    desc="Drive 1",
    comment="Drive line 1 on qubit 1",
    connected=["Q1"],
    hamiltonian_func=hamiltonians.x_drive,
)
drive2 = Drive(
    name="d2",
    desc="Drive 2",
    comment="Drive line 2 on qubit 2",
    connected=["Q2"],
    hamiltonian_func=hamiltonians.x_drive,
)

m00_q1 = 0.97  # Prop to read qubit 1 state 0 as 0
m01_q1 = 0.04  # Prop to read qubit 1 state 0 as 1
m00_q2 = 0.96  # Prop to read qubit 2 state 0 as 0
m01_q2 = 0.05  # Prop to read qubit 2 state 0 as 1
one_zeros = np.array([0] * qubit_lvls)
zero_ones = np.array([1] * qubit_lvls)
one_zeros[0] = 1
zero_ones[0] = 0
val1 = one_zeros * m00_q1 + zero_ones * m01_q1
val2 = one_zeros * m00_q2 + zero_ones * m01_q2
min = one_zeros * 0.8 + zero_ones * 0.0
max = one_zeros * 1.0 + zero_ones * 0.2
confusion_row1 = Quantity(value=val1, min_val=min, max_val=max, unit="")
confusion_row2 = Quantity(value=val2, min_val=min, max_val=max, unit="")
conf_matrix = ConfusionMatrix(Q1=confusion_row1, Q2=confusion_row2)

init_temp = 50e-3
init_ground = InitialiseGround(
    init_temp=Quantity(value=init_temp, min_val=-0.001, max_val=0.22, unit="K")
)

model = Model(
    [q1, q2],  # Individual, self-contained components
    [drive, drive2, q1q2],  # Interactions between components
    [conf_matrix, init_ground],  # SPAM processing
)

pmap = ParameterMap(model=model)
model.set_dressed(False)

hdrift, hks = model.get_Hamiltonians()

with open("test/model.pickle", "rb") as filename:
    test_data = pickle.load(filename)


@pytest.mark.unit
def test_model_eigenfrequencies_1() -> None:
    "Eigenfrequency of qubit 1"
    assert hdrift[3, 3] - hdrift[0, 0] == freq_q1 * 2 * np.pi


@pytest.mark.unit
def test_model_eigenfrequencies_2() -> None:
    "Eigenfrequency of qubit 2"
    assert hdrift[1, 1] - hdrift[0, 0] == freq_q2 * 2 * np.pi


@pytest.mark.unit
def test_model_couplings() -> None:
    assert hks["d1"][3, 0] == 1
    assert hks["d2"][1, 0] == 1


@pytest.mark.unit
def test_model_get_hamiltonian() -> None:
    ham = model.get_Hamiltonian()
    np.testing.assert_allclose(ham, hdrift)

    sig = {"d1": {"ts": np.linspace(0, 5e-9, 10), "values": np.linspace(0e9, 20e9, 10)}}
    hams = model.get_Hamiltonian(sig)
    np.testing.assert_allclose(hams, test_data["sliced_hamiltonians"])


@pytest.mark.unit
def test_get_qubit_frequency() -> None:
    np.testing.assert_allclose(
        model.get_qubit_freqs(), [4999294802.027272, 5600626454.433859]
    )


@pytest.mark.unit
def test_get_indeces() -> None:
    assert model.get_state_index((0, 0)) == 0
    assert model.get_state_index((1, 0)) == 3
    assert model.get_state_index((1, 1)) == 4
    assert model.get_state_index((2, 1)) == 7

    actual = model.get_state_indeces([(0, 0), (1, 0), (2, 0), (1, 1)])
    desired = [0, 3, 6, 4]
    np.testing.assert_equal(actual=actual, desired=desired)


@pytest.mark.unit
def test_model_update_by_parametermap() -> None:
    pmap.set_parameters([freq_q1 * 0.9995], [[("Q1", "freq")]])
    hdrift_a, _ = model.get_Hamiltonians()

    pmap.set_parameters([freq_q1 * 1.0005], [[("Q1", "freq")]])
    hdrift_b, _ = model.get_Hamiltonians()

    assert hdrift_a[3, 3] - hdrift_a[0, 0] == freq_q1 * 0.9995 * 2 * np.pi
    assert hdrift_b[3, 3] - hdrift_b[0, 0] == freq_q1 * 1.0005 * 2 * np.pi
